# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    src/wenv/_core/path.py: Wine-Python environment paths

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import os

from .pythonversion import PythonVersion
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# PATHS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Paths:
    """
    Wine Python environment paths
    """

    def __init__(self, pythonprefix: str, pythonversion: PythonVersion):

        self._pythonprefix = pythonprefix
        self._pythonversion_block = pythonversion.as_block()

    def __getitem__(self, key: str) -> str:

        if key == "pythonprefix":
            return self._pythonprefix
        if key == "lib":
            return os.path.join(self["pythonprefix"], "Lib")
        if key == "sitepackages":
            return os.path.join(self["lib"], "site-packages")
        if key == "sitecustomize":
            return os.path.join(self["sitepackages"], "sitecustomize.py")
        if key == "scripts":
            return os.path.join(self["pythonprefix"], "Scripts")
        if key == "interpreter":
            return os.path.join(self["pythonprefix"], "python.exe")
        if key == "pip":
            return os.path.join(self["scripts"], "pip.exe")
        if key == "libzip":
            return os.path.join(
                self["pythonprefix"], "python%s.zip" % self._pythonversion_block
            )
        if key == "pth":
            return os.path.join(
                self["pythonprefix"], "python%s._pth" % self._pythonversion_block
            )

        raise KeyError("not a valid path key")

    @staticmethod
    def symlink(src, dest: str):
        """
        Generates a symlink and checks result
        """

        if not os.path.lexists(dest):
            os.symlink(src, dest)

        if not os.path.exists(dest):
            raise OSError('"{LINK:s}" could not be created'.format(LINK=dest))
        if not os.path.islink(dest):
            raise OSError('"{LINK:s}" is not a symlink'.format(LINK=dest))
        if os.readlink(dest) != src:
            raise OSError('"{LINK:s}" points to the wrong source'.format(LINK=dest))



# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    tests/lib/const.py: Holds constant values, flags, types

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CONST
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ARCHS = ["win32", "win64"]
DEFAULT_TIMEOUT = 600



# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    tests/test_init.py: Testing clean & init

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from .lib import get_context, run_process, no_errors_in, remove_colors

import pytest

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# TEST(s)
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@pytest.mark.parametrize("arch,build", get_context())
def test_1_clean(arch, build):

    out, err, code = run_process(["wenv", "clean"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})

    assert code == 0
    assert no_errors_in(err)


@pytest.mark.parametrize("arch,build", get_context())
def test_2_init(arch, build):

    out, err, code = run_process(["wenv", "help"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})

    assert code == 0
    assert no_errors_in(err)
    assert "wenv pip" not in out
    assert 'wenv python' not in remove_colors(out)

    out, err, code = run_process(["wenv", "init"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})

    assert code == 0
    # assert len(err.strip()) == 0 # pip output goes to stderr

    out, err, code = run_process(["wenv", "help"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})

    assert code == 0
    assert no_errors_in(err)
    assert "wenv pip" in out
    assert 'wenv python' in remove_colors(out)


@pytest.mark.parametrize("arch,build", get_context())
def test_3_clean(arch, build):

    out, err, code = run_process(["wenv", "clean"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})

    assert code == 0
    assert no_errors_in(err)


@pytest.mark.parametrize("arch,build", get_context())
def test_4_offline(arch, build):

    out, err, code = run_process(["wenv", "help"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})

    assert code == 0
    assert no_errors_in(err)
    assert "wenv pip" not in out
    assert 'wenv python' not in remove_colors(out)

    out, err, code = run_process(["wenv", "cache"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})

    assert code == 0
    assert no_errors_in(err)
    assert "wenv pip" not in out

    out, err, code = run_process(
        ["wenv", "init"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build), "WENV_OFFLINE": "true"}
    )

    assert code == 0
    # assert len(err.strip()) == 0 # pip output goes to stderr

    out, err, code = run_process(["wenv", "help"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})

    assert code == 0
    assert no_errors_in(err)
    assert "wenv pip" in out
    assert 'wenv python' in remove_colors(out)



# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    src/wenv/_core/pythonversion.py: Parse and handle Python versions

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Any, Union

try:
    from typing import NotImplementedType # re-introduced in Python 3.10
except ImportError:
    NotImplementedType = type(NotImplemented)

from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class PythonVersion:
    """
    Parse, compare and export Python versions. Immutable.

    Args:
        arch : Build architecture. Can be ``win32``, ``win64`` or ``arm64``.
        major : Python major version, i.e. ``X`` from ``X.0.0``.
        minor : Python minor version, i.e. ``X`` from ``0.X.0``.
        maintenance : Python maintenance version, i.e. ``X`` from ``0.0.X``.
        build : Type of build, e.g. ``aX``, ``bX``, ``rcX``, etc. If left empty or set to ``stable``, the build is considered stable.
    """

    def __init__(self, arch: str, major: int, minor: int, maintenance: int, build: str = "stable"):

        if arch not in ("win32", "win64", "arm64"):
            raise ValueError("Unknown arch: " + arch)
        if major <= 2:
            raise ValueError("Only Python 3 and newer supported")

        self._arch = arch
        self._major, self._minor, self._maintenance = major, minor, maintenance
        self._build = "stable" if build == "" else build

    def __str__(self) -> str:
        """
        Converts version to string.
        """

        return "%d.%d.%d.%s" % (
            self._major,
            self._minor,
            self._maintenance,
            self._build,
        )

    def __repr__(self) -> str:
        """
        Converts version to interactive string representation.
        """

        return "<Python %d.%d.%d.%s (%s)>" % (
            self._major,
            self._minor,
            self._maintenance,
            self._build,
            self._arch,
        )

    def __eq__(self, other: Any) -> Union[bool, NotImplementedType]:
        """
        Equality operator.
        """

        if not isinstance(other, type(self)):
            return NotImplemented

        return self._as_sort() == other._as_sort()

    def __gt__(self, other: Any) -> Union[bool, NotImplementedType]:
        """
        Greater than operator.
        """

        if not isinstance(other, type(self)):
            return NotImplemented

        return self._as_sort() > other._as_sort()

    def __lt__(self, other: Any) -> Union[bool, NotImplementedType]:
        """
        Lesser than operator.
        """

        if not isinstance(other, type(self)):
            return NotImplemented

        return self._as_sort() < other._as_sort()

    def _as_sort(self) -> str:

        return "%04d-%04d-%04d-%s" % (
            self._major,
            self._minor,
            self._maintenance,
            self._build,
        )

    @property
    def arch(self) -> str:
        """
        Build architecture.
        """

        return self._arch

    @property
    def major(self) -> int:
        """
        Python major version.
        """

        return self._major

    @property
    def minor(self) -> int:
        """
        Python minor version.
        """

        return self._minor

    @property
    def maintenance(self) -> int:
        """
        Python maintenance version.
        """

        return self._maintenance

    @property
    def build(self) -> str:
        """
        Type of build.
        """

        return self._build

    def as_block(self) -> str:
        """
        Returns:
            "Block", i.e. combination of major and minor version as string.
        """

        return "%d%d" % (self._major, self._minor)

    def as_config(self) -> str:
        """
        Returns:
            String for use in configuration.
        """

        return str(self)

    def as_url(self) -> str:
        """
        Returns:
            Download URL for Windows Embedded Build ZIP file.
        """

        return (
            "https://www.python.org/ftp/python/%d.%d.%d/"
            % (self._major, self._minor, self._maintenance)
            + self.as_zipname()
        )

    def as_zipname(self) -> str:
        """
        Returns:
            Name of Windows Embedded Build ZIP file.
        """

        build = "" if self._build == "stable" else self._build
        arch = "amd64" if self._arch == "win64" else self._arch
        sub_tuple = (self._major, self._minor, self._maintenance, build, arch)

        if build.startswith("post"):
            return "python-%d.%d.%d.%s-embed-%s.zip" % sub_tuple
        else:
            return "python-%d.%d.%d%s-embed-%s.zip" % sub_tuple

    @classmethod
    def from_config(cls, arch: str, version: str):
        """
        Parses version from configuration value.

        Args:
            arch : Build architecture.
            version : Full version string, including type of build.
        """

        segments = version.split(".")
        if not len(segments) in (3, 4):
            raise ValueError("wrong number of version segments")
        if len(segments) == 3:
            segments.append("stable")
        if not all((segment.isnumeric() for segment in segments[:3])):
            raise ValueError("version segments are not numeric")
        segments = tuple([int(segment) for segment in segments[:3]] + [segments[3]])

        return cls(arch, *segments)

    @classmethod
    def from_zipname(cls, zip_name: str):
        """
        Parse version from Windows Embedded Build ZIP file name.

        Args:
            zip_name : Name of file
        """

        if not isinstance(zip_name, str):
            raise TypeError("zip_name must be str")

        fragments = zip_name.split("-")
        fragments.append(fragments[3].split(".")[1])
        fragments[3] = fragments[3].split(".")[0]

        if not fragments[0] == "python":
            raise ValueError('fagment[0] != "python"')
        if not fragments[2] == "embed":
            raise ValueError('fagment[2] != "embed"')
        if not fragments[3] in ("win32", "amd64", "arm64"):
            raise ValueError('fagment[3] not in in ("win32", "amd64", "arm64")')
        if not fragments[4] == "zip":
            raise ValueError('fagment[4] != "zip"')

        arch = "win64" if fragments[3] == "amd64" else fragments[3]
        release = [int(f) if f.isnumeric() else f for f in fragments[1].split(".")]

        if isinstance(release[2], str):
            if not len(release[2]) > 0:
                raise ValueError("broken maintenance/build fragment")
            pos = None
            for pos_idx, char in enumerate(release[2]):
                if not char.isdigit():
                    pos = pos_idx
                    break
            if pos is None:
                raise ValueError("maintenance fragment could not be identified")
            release.append(release[2][pos:])
            release[2] = release[2][:pos]
            if not release[2].isnumeric():
                raise ValueError("maintenance fragment not numeric")
            release[2] = int(release[2])

        if len(release) == 3:
            release.append("stable")
        if not len(release) == 4:
            raise ValueError(
                "release does not have 4 fragments (major,minor,maintenance,build)"
            )

        return cls(arch, *release)



# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    tests/test_shebang.py: Testing Python interpreter shebang alias

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from .lib import get_context, run_process, no_errors_in

import pytest

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# TEST(s)
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


def test_without_shebang():

    out, err, code = run_process(["python", "tests/shebang.py"])

    assert code == 0
    assert no_errors_in(err)
    assert "Hello from Linux on x86_64!" == out.strip()


@pytest.mark.parametrize("arch,build", get_context())
def test_with_shebang(arch, build):

    out, err, code = run_process(
        ["./tests/shebang.py"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)}
    )

    assert code == 0
    assert no_errors_in(err)
    if arch == 'win32':
        assert "Hello from Windows on x86!" == out.strip()
    else:
        assert "Hello from Windows on AMD64!" == out.strip()



# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    src/wenv/_core/env.py: Managing a Wine-Python environment

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import json
from io import BytesIO
import os
import shutil
import subprocess
import sys
from typing import Any, Dict, Generator, List
import zipfile

from .config import EnvConfig
from .const import c, COVERAGE_STARTUP, HELP_STR
from .paths import Paths
from .source import download
from .typeguard import typechecked

import wenv # HACK für version

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Env:
    """
    Represents one Wine Python environment. Mutable.

    args:
        kwargs : An arbitrary number of keyword arguments matching valid configuration options. In previous releases, the constructor expected one optional argument, ``parameter``. It should either be ``None`` or a dictionary. In the latter case, the dictionary may contain all valid configuration options. ``parameter`` can still be used but is deprecated.
    """

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # INIT
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def __init__(self, **kwargs: Any):

        if "parameter" in kwargs.keys():  # legacy API
            if len(kwargs) > 1:
                raise ValueError(
                    'legacy API: only allows one parameter dict, named "parameter"'
                )
            kwargs = kwargs["parameter"]
            if kwargs is None:
                kwargs = dict()
            if not isinstance(kwargs, dict) and not isinstance(kwargs, EnvConfig):
                raise TypeError(
                    "legacy API: only allows one parameter, must be a dict or EnvConfig"
                )
            if not isinstance(kwargs, EnvConfig):
                kwargs = EnvConfig(**kwargs)
        else:
            kwargs = EnvConfig(**kwargs)

        self._p = kwargs
        self._cmd_dict = None
        self._cli_dict = None
        self._envvar_dict = None

        self._init_dicts()

    def _init_dicts(self):
        """
        Initialize core dictionaries. Function can also be used for re-initialization.
        """

        # Init Wine cmd names
        self._wine_dict = {
            "win32": self._p["wine_bin_win32"],
            "win64": self._p["wine_bin_win64"],
            "arm64": self._p["wine_bin_arm64"],
        }

        # Init Python environment paths
        self._path_dict = Paths(self._p["pythonprefix"], self._p["pythonversion"])
        # Init Python commands and scripts
        self._init_cmd_dict()
        # Init internal CLI commands
        self._init_cli_dict()
        # Init environment variables
        self._init_envvar_dict()

    def _init_cmd_dict(self):

        @typechecked
        def ls_exe(directory: str) -> Generator:
            if not os.path.isdir(directory):
                return
            for item in os.listdir(directory):
                if not item.lower().endswith(".exe"):
                    continue
                yield item[:-4], os.path.join(directory, item)

        self._cmd_dict = dict(ls_exe(self._path_dict["scripts"]))
        self._cmd_dict.update(dict(ls_exe(self._path_dict["pythonprefix"])))

    def _init_cli_dict(self):

        self._cli_dict = {
            item[5:]: getattr(self, item)
            for item in dir(self)
            if item.startswith("_cli_") and hasattr(getattr(self, item), "__call__")
        }

    def _init_envvar_dict(self):

        self._envvar_dict = os.environ.copy()
        self._envvar_dict.update(
            dict(
                WINEARCH=self._p["arch"],  # Architecture
                WINEPREFIX=self._p["wineprefix"],  # Wine prefix / directory
                WINEDLLOVERRIDES="mscoree=d",  # Disable MONO: https://unix.stackexchange.com/a/191609
                WINEDEBUG=self._p["winedebug"],  # Wine debug level
                PYTHONHOME=self._p[
                    "pythonprefix"
                ],  # Python home for Wine Python (can be a Unix path)
                VIRTUAL_ENV="",  # Reset Unix virtual env variable - wenv is "independent"
                PIP_NO_WARN_SCRIPT_LOCATION="0",  # pip will not warn that pythonprefix and scripts are not in PATH
            )
        )
        if self._p["wineinstallprefix"] not in (
            None,
            "",
        ):  # allow custom installations of Wine outside of PATH
            path = self._envvar_dict.get("PATH", "")
            self._envvar_dict["PATH"] = (
                os.path.join(self._p["wineinstallprefix"], "bin") + ":" + path
            )
            ld_library_path = self._envvar_dict.get("LD_LIBRARY_PATH", "")
            self._envvar_dict["LD_LIBRARY_PATH"] = ":".join(
                (
                    os.path.join(self._p["wineinstallprefix"], "lib"),
                    os.path.join(self._p["wineinstallprefix"], "lib64"),
                    ld_library_path,
                )
            )  # https://wiki.winehq.org/FAQ#Can_I_install_more_than_one_Wine_version_on_my_system.3F

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # WINE BUG #47766 / ZUGBRUECKE BUG #49
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def wine_47766_workaround(self):
        """
        Due to `Wine bug #47766`_ (in ``PathAllocCanonicalize``), *Wine* crashes if **any** folder in the path to ``pythonprefix`` is hidden Unix-style (i.e. prefixed with a dot / ``.``). This workaround creates a symlink directly pointing to ``pythonprefix`` into ``/tmp``, which is (more or less) guaranteed to be visible. It is then used instead of the actual ``pythonprefix``.

        Run ``setup_wineprefix`` and ``setup_pythonprefix`` **before** calling ``wine_47766_workaround``. Any subsequent action such as installing or using ``pip`` must happen **after** calling ``wine_47766_workaround``.

        .. _Wine bug #47766: https://bugs.winehq.org/show_bug.cgi?id=47766
        """

        is_clean = lambda path: not any(
            [seg.startswith(".") for seg in path.split(os.path.sep)]
        )

        pythonprefix = os.path.abspath(self._p["pythonprefix"])

        if pythonprefix != self._p["pythonprefix"]:
            self._p["pythonprefix"] = pythonprefix
            self._init_dicts()

        if is_clean(self._p["pythonprefix"]):
            return

        import tempfile, hashlib

        link_path = os.path.join(
            tempfile.gettempdir(),
            "wenv-"
            + hashlib.sha256(self._p["pythonprefix"].encode("utf-8")).hexdigest()[:8],
        )
        if not is_clean(link_path):
            raise OSError(
                'unable to create clean link path: "{LINK:s}"'.format(LINK=link_path)
            )

        if os.path.exists(self._p["pythonprefix"]):
            Paths.symlink(self._p["pythonprefix"], link_path)

        self._p["pythonprefix"] = link_path
        self._init_dicts()

    def wine_47766_workaround_uninstall(self):
        """
        Reverts the `Wine bug #47766`_ workaround, i.e. it removes the symlink.
        """

        self.wine_47766_workaround()

        if os.path.lexists(self._p["pythonprefix"]):
            os.unlink(self._p["pythonprefix"])

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # ENSURE ENVIRONMENT
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def ensure(self):
        """
        Equivalent to ``wenv init``. Intended to be used by 3rd-party packages which want to "ensure" that ``wenv`` has been initialized (i.e. *Python* and *pip* are present and working). ``ensure()`` calls the following methods:

            - :meth:`wenv.Env.setup_wineprefix`
            - :meth:`wenv.Env.setup_pythonprefix`
            - :meth:`wenv.Env.wine_47766_workaround`
            - :meth:`wenv.Env.setup_pip`
        """

        self.setup_wineprefix()
        self.setup_pythonprefix()
        self.wine_47766_workaround()  # must run after setup_pythonprefix and before setup_pip
        self.setup_pip()

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # DESTROY / UNINSTALL ENVIRONMENT
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def uninstall(self):
        """
        Equivalent to ``wenv clean``. It removes the current Wine Python environment, i.e. Python interpreter, pip, setuptools, wheel and all installed packages.
        """

        # Does Wine prefix exist?
        if os.path.exists(self._p["wineprefix"]):
            # Delete tree
            shutil.rmtree(self._p["wineprefix"])

        # Does Python prefix exist?
        if os.path.exists(self._p["pythonprefix"]):
            # Delete tree
            shutil.rmtree(self._p["pythonprefix"])

        self.wine_47766_workaround_uninstall()

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # CACHE INSTALLATION FILES LOCALLY
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def cache(self):
        """
        Equivalent to ``wenv cache``. It fetches installation files and caches them for offline usage, including the Python interpreter, pip, setuptools and wheel.
        """

        os.makedirs(self._p["cache"], exist_ok=True)

        with open(os.path.join(self._p["cache"], self._p["pythonversion"].as_zipname()), "wb") as f:
            f.write(self._get_python(offline=False))
        with open(os.path.join(self._p["cache"], "get-pip.py"), "wb") as f:
            f.write(self._get_pip(offline=False))

        for package in ("pip", "setuptools", "wheel"):
            self.cache_package(package)

    def cache_package(self, name: str):
        """
        Caches a specific package by nameself.

        Args:
            name : Name of PyPI package
        """

        os.makedirs(self._p["packages"], exist_ok=True)

        meta = json.loads(
            download("https://pypi.org/pypi/%s/json" % name, mode="binary").decode(
                "utf-8"
            )
        )

        for item in meta["urls"]:
            with open(os.path.join(self._p["packages"], item["filename"]), "wb") as f:
                f.write(download(item["url"], mode="binary"))

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Fetch installer data
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def _get_python(self, offline: bool = False) -> bytes:

        if offline:
            with open(os.path.join(self._p["cache"], self._p["pythonversion"].as_zipname()), "rb") as f:
                return f.read()
        else:
            return download(self._p["pythonversion"].as_url(), mode="binary")

    def _get_pip(self, offline: bool = False) -> bytes:

        if offline:
            with open(os.path.join(self._p["cache"], "get-pip.py"), "rb") as f:
                return f.read()
        else:
            return download("https://bootstrap.pypa.io/get-pip.py", mode="text").encode(
                "utf-8"
            )

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # SETUP
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def setup_wineprefix(self, overwrite: bool = False):
        """
        Part of the initialization process, but can be triggered on its own if required. It creates a *Wine* prefix according to *wenv*'s configuration.

        Args:
            overwrite : If set to ``True``, a pre-existing *Wine* prefix is removed before a new one is created.
        """

        if not isinstance(overwrite, bool):
            raise TypeError("overwrite is not a boolean")

        # Does it exist?
        if os.path.exists(self._p["wineprefix"]):
            # Exit if overwrite flag is not set
            if not overwrite:
                return
            # Delete if overwrite is set
            shutil.rmtree(self._p["wineprefix"])

        os.makedirs(
            self._p["wineprefix"]
        )  # HACK Wine on Travis CI expects folder to exist

        # Start wine server into prepared environment
        envvar_dict = self._envvar_dict.copy()
        envvar_dict["DISPLAY"] = ""
        proc = subprocess.Popen(["wine", "wineboot", "-i"], env=envvar_dict)
        proc.wait()
        if proc.returncode != 0:
            sys.exit(1)

        self._init_dicts()

    def setup_pythonprefix(self, overwrite: bool = False):
        """
        Part of the initialization process, but can be triggered on its own if required. It installs the *CPython* interpreter into the *Python* prefix.

        Args:
            overwrite : If set to ``True``, a pre-existing *Python* prefix is removed before a new one is created.
        """

        if not isinstance(overwrite, bool):
            raise TypeError("overwrite is not a boolean")

        # Is there a pre-existing Python installation with identical parameters?
        preexisting = os.path.isfile(self._path_dict["interpreter"])

        # Is there a preexisting installation and should it be overwritten?
        if preexisting and overwrite:
            # Delete folder
            shutil.rmtree(self._p["pythonprefix"])

        # Make sure the target directory exists
        if not os.path.exists(self._p["pythonprefix"]):
            # Create folder
            os.makedirs(self._p["pythonprefix"])

        # Only do if Python is not there OR if should be overwritten
        if overwrite or not preexisting:

            # Generate in-memory file-like-object
            archive_zip = BytesIO()
            # Fetch Python zip file
            archive_zip.write(self._get_python(self._p["offline"]))
            # Unpack from memory to disk
            with zipfile.ZipFile(archive_zip) as f:
                f.extractall(
                    path=self._p["pythonprefix"]
                )  # Directory created if required

            # Unpack Python library from embedded zip on disk
            with zipfile.ZipFile(self._path_dict["libzip"], "r") as f:
                f.extractall(
                    path=self._path_dict["lib"]
                )  # Directory created if required
            # Remove Python library zip from disk
            os.remove(self._path_dict["libzip"])

            # HACK: Fix library path in pth-file (CPython >= 3.6)
            with open(self._path_dict["pth"], "w") as f:
                f.write(
                    "Lib\n.\n\n# Uncomment to run site.main() automatically\nimport site\n"
                )

        # Create site-packages folder if it does not exist
        if not os.path.exists(self._path_dict["sitepackages"]):
            # Create folder
            os.makedirs(self._path_dict["sitepackages"])

        self._init_dicts()

    def setup_pip(self):
        """
        Part of the initialization process, but can be triggered on its own if required. It installs ``pip``, assuming that both the ``wineprefix`` and ``pythonprefix`` are already present.
        """

        # Exit if it exists
        if os.path.isfile(self._path_dict["pip"]):
            return

        envvar_dict = {k: os.environ[k] for k in os.environ.keys()}
        envvar_dict.update(self._p.export_envvar_dict())

        if self._p["offline"]:
            proc = subprocess.Popen(
                [
                    "wenv",
                    "python",
                    os.path.join(self._p["cache"], "get-pip.py"),
                    "--no-index",
                    "--find-links=%s" % self._p["packages"],
                ],
                env=envvar_dict,
            )
            proc.wait()
        else:
            getpip = self._get_pip(self._p["offline"])
            proc = subprocess.Popen(
                ["wenv", "python"], stdin=subprocess.PIPE, env=envvar_dict
            )
            proc.communicate(input=getpip)

        self._init_dicts()

    def setup_coverage_activate(self):
        """
        Equivalent to ``wenv init_coverage``. It enables coverage analysis inside wenv.
        """

        # Ensure that coverage is started with the Python interpreter
        siteconfig_cnt = ""
        if os.path.isfile(self._path_dict["sitecustomize"]):
            with open(self._path_dict["sitecustomize"], "r") as f:
                siteconfig_cnt = f.read()
            if COVERAGE_STARTUP in siteconfig_cnt:
                return
        with open(self._path_dict["sitecustomize"], "w") as f:
            f.write(siteconfig_cnt + "\n" + COVERAGE_STARTUP)

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # PACKAGE MANAGEMENT
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def install_package(self, name: str, update: bool = False):
        """
        Thin wrapper for ``wenv pip install {-U} {name}``. Installs and/or updates a package.

        Args:
            name : Name of PyPI package
            update : Update flag
        """

        if not isinstance(name, str):
            raise TypeError("name must be str")
        if len(name) == 0:
            raise ValueError("name must not be empty")
        if not isinstance(update, bool):
            raise TypeError("update must be bool")

        cmd = ["wenv", "pip", "install"]
        if update:
            cmd.append("-U")
        cmd.append(name)

        envvar_dict = {k: os.environ[k] for k in os.environ.keys()}
        envvar_dict.update(self._p.export_envvar_dict())

        proc = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=envvar_dict
        )
        outs, errs = proc.communicate()
        if proc.returncode != 0:
            raise SystemError('installing package "%s" failed' % name, outs, errs)

        self._init_dicts()

    def uninstall_package(self, name: str):
        """
        Thin wrapper for ``wenv pip uninstall -y {name}``. Removes a package.

        Args:
            name : Name of PyPI package
        """

        if not isinstance(name, str):
            raise TypeError("name must be str")
        if len(name) == 0:
            raise ValueError("name must not be empty")

        envvar_dict = {k: os.environ[k] for k in os.environ.keys()}
        envvar_dict.update(self._p.export_envvar_dict())

        proc = subprocess.Popen(
            ["wenv", "pip", "uninstall", "-y", name],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=envvar_dict,
        )
        outs, errs = proc.communicate()
        if proc.returncode != 0:
            raise SystemError('uninstalling package "%s" failed' % name, outs, errs)

        self._init_dicts()

    def list_packages(self) -> List[Dict[str, str]]:
        """
        Thin wrapper for ``wenv pip list --format json``.

        Returns:
            A list of dictionaries of format ``{"name": "Name of PyPI package ", "version": "package version"}``.
        """

        envvar_dict = {k: os.environ[k] for k in os.environ.keys()}
        envvar_dict.update(self._p.export_envvar_dict())

        proc = subprocess.Popen(
            ["wenv", "pip", "list", "--format", "json"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=envvar_dict,
        )
        outs, errs = proc.communicate()
        if proc.returncode != 0:
            raise SystemError("listing packages failed", outs, errs)

        return json.loads(outs.decode("utf-8"))

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # CLI
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def _cli_init(self):
        "sets up an environment (Wine prefix, Python interpreter, pip, setuptools, wheel)"

        self.ensure()

    def _cli_init_coverage(self):
        "enables coverage analysis inside wenv"

        self.setup_coverage_activate()

    def _cli_cache(self):
        "fetches installation files and caches them for offline usage (Python interpreter, pip, setuptools, wheel)"

        self.cache()

    def _cli_clean(self):
        "removes current environment (Python interpreter, pip, setuptools, wheel, all installed packages)"

        self.uninstall()

    def _cli_help(self):
        "prints this help text"

        def colorize(text):
            for color, key in (
                ("GREEN", "python"),
                ("CYAN", "pip"),
                ("YELLOW", "wheel"),
                ("MAGENTA", "pytest"),
                ("MAGENTA", "py.test"),
            ):
                text = text.replace(key, c[color] + key + c["RESET"])
            return text

        scripts = colorize(
            "\n".join(
                [
                    "- wenv {SCRIPT:s}".format(
                        SCRIPT=key,
                    )
                    for key in sorted(self._cmd_dict.keys())
                ]
            )
        )
        if len(scripts) == 0:
            scripts = "(None)"

        sys.stdout.write(
            HELP_STR.format(
                VERSION=wenv.__version__,
                CLIS="\n".join(
                    [
                        "- wenv {CLI:s}: {HELP:s}".format(
                            CLI=key,
                            HELP=self._cli_dict[key].__doc__,
                        )
                        for key in sorted(self._cli_dict.keys())
                    ]
                ),
                SCRIPTS=scripts,
            )
        )
        sys.stdout.flush()

    def _cli_version(self):
        "shows wenv's version (also available as `--version`)"

        sys.stdout.write(wenv.__version__ + '\n')
        sys.stdout.flush()

    def cli(self):
        """
        Command line interface entry point. Equivalent to ``wenv [...]``. Looks for sub-commands and parameters in ``sys.argv``.
        """

        # No command passed
        if len(sys.argv) < 2:
            sys.stderr.write("There was no command passed.\n")
            sys.stderr.flush()
            sys.exit(1)

        # Separate command and arguments
        cmd, param = sys.argv[1], sys.argv[2:]

        # Allow -h and --help
        if cmd in ["-h", "--help"]:
            cmd = "help"

        # Version exposed
        if cmd == "--version":
            cmd = "version"

        # Special CLI command
        if cmd in self._cli_dict.keys():
            self._cli_dict[cmd]()
            sys.exit(0)

        # Command is unknown
        if cmd not in self._cmd_dict.keys():
            sys.stderr.write('Unknown command or script: "{CMD:s}"\n'.format(CMD=cmd))
            sys.stderr.flush()
            sys.exit(1)

        # Get Wine depending on arch
        wine = self._wine_dict[self._p["arch"]]

        self.wine_47766_workaround()

        # Replace this process with Wine
        os.execvpe(
            wine,
            (wine, self._cmd_dict[cmd])
            + tuple(param),  # Python 3.4: No in-place unpacking of param
            self._envvar_dict,
        )

    def shebang(self):
        """
        shebang entry point for Wine Python interpreter. Equivalent to ``_wenv_python [...]``. Does not look at ``sys.argv``. It only passes ``sys.argv[1]`` on to the Wine Python interpreter.

        This interface is working around a lack of Unix specification, see:

            - https://stackoverflow.com/q/4303128/1672565
            - https://unix.stackexchange.com/q/63979/28301
            - https://lists.gnu.org/archive/html/bug-sh-utils/2002-04/msg00020.html
        """

        if len(sys.argv) < 2:
            raise OSError(
                "entry point meant to be used as a shebang but no file name was provided"
            )

        # Get Wine depending on arch
        wine = self._wine_dict[self._p["arch"]]

        self.wine_47766_workaround()

        # Replace this process with Wine
        os.execvpe(
            wine, (wine, self._cmd_dict["python"], sys.argv[1]), self._envvar_dict
        )


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLI EXPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


def cli():

    Env().cli()


def shebang():

    Env().shebang()



# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    src/wenv/_core/source.py: Obtaining Python and pip locally or remotely

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Optional, Union
from urllib.request import urlopen, Request

from .pythonversion import PythonVersion
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
def download(down_url: str, mode: str = "binary") -> Union[str, bytes]:

    assert mode in ("text", "binary")
    assert isinstance(down_url, str)

    httprequest = Request(down_url)

    with urlopen(httprequest) as response:
        assert response.status == 200
        data = response.read()

    if mode == 'text':
        return data.decode('utf-8')

    return data # mode == 'binary'


@typechecked
def get_latest_python_build(
    arch: str,
    major: int,
    minor: int,
    builds: Optional[List[PythonVersion]] = None,
) -> Optional[PythonVersion]:
    """
    Find the latest build of a given Python major and minor version for a given architecture.
    Returns ``None`` if none can be found.

    Args:
        arch : Build architecture.
        major : Python major version.
        minor : Python minor version.
        builds : A list of :class:`wenv.PythonVersion` objects. If left empty, ``python.org`` will be queried.
    Returns:
        A :class:`wenv.PythonVersion` object or ``None``.
    """

    if builds is None:
        builds = get_available_python_builds()

    filtered_build = [
        build
        for build in builds
        if all((
            build.arch == arch,
            build.major == major,
            build.minor == minor,
        ))
    ]

    if len(filtered_build) == 0:
        return None

    filtered_build.sort()
    return filtered_build[-1]


@typechecked
def get_available_python_builds(parallel: int = 8) -> List[PythonVersion]:
    """
    Queries ``python.org`` for Windows Embedded Builds.

    Args:
        parallel : Number of parallel queries to ``python.org``.
    Returns:
        All available Windows Embedded Builds of CPython 3.
    """

    versions = [
        tuple(int(nr) for nr in line.split('"')[1][:-1].split("."))
        for line in download("https://www.python.org/ftp/python/", mode="text").split(
            "\n"
        )
        if all(
            (
                line.startswith('<a href="'),
                line[9:10].isdigit(),
                not line.startswith('<a href="2.'),
            )
        )
    ] # get URLs of all verions, dump Python 2

    version_urls = [
        "https://www.python.org/ftp/python/%d.%d.%d/" % version
        if len(version) == 3
        else "https://www.python.org/ftp/python/%d.%d/" % version
        for version in versions
    ] # some URLs in early Python 3.X releases are following  `major.minor` pattern

    with ThreadPoolExecutor(max_workers = parallel) as p:
        version_futures = [
            p.submit(
                lambda x: download(x, mode="text"), version_url
            ) for version_url in version_urls
        ] # get inventory of all maintenance version downloads
        version_downloads = [
            future.result()
            for future in as_completed(version_futures)
        ]

    return [
        PythonVersion.from_zipname(line.split('"')[1])
        for version, version_url, version_download in zip(
            versions, version_urls, version_downloads
        )
        for line in version_download.split("\n")
        if all(
            (
                line.startswith('<a href="'),
                "embed" in line,
                ".zip" in line,
                ".zip.asc" not in line,
            )
        )
    ]



# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    tests/lib/output.py: Looking at output

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import re

from .param import get_context, run_process


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def no_errors_in(output: str) -> bool:

    lines = output.split('\n')

    for line in lines:

        line = line.strip()

        if len(line) == 0:
            continue

        if all(
            fragment in line.lower()
            for fragment in (
                'debug mode',
                'run-time type checks',
                'runtimewarning',
            )
        ):
            continue

        if 'wine client error:' in line:
            print('IGNORED:', line)
            continue

        return False

    return True


def remove_colors(output: str) -> str:

    # https://stackoverflow.com/a/14693789/1672565
    return re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])').sub('', output)



# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    src/wenv/_core/config.py: Handles the modules configuration

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import os
import json
import site
import sys
from typing import Any, Dict, Optional

from .const import CONFIG_FN
from .errors import EnvConfigParserError
from .pythonversion import PythonVersion
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CONFIGURATION CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class EnvConfig(dict):
    """
    Wine Python environment configuration. Subclass of ``dict``.
    It holds default values and overwrites them with values found in configuration files and environment variables.
    Usually, one should not work with this class directly - unless an access to current configuration parameters is desired.

    Args:
        override : Specify custom values for configuration parameters via keyword arguments.
    """

    _KEYS = (
        "arch",
        "pythonversion",
        "wine_bin_win32",
        "wine_bin_win64",
        "wine_bin_arm64",
        "winedebug",
        "wineinstallprefix",
        "prefix",
        "wineprefix",
        "pythonprefix",
        "offline",
        "cache",
        "packages",
        "_issues_50_workaround",
    )

    def __init__(self, **override: Any):

        # Call parent constructur, just in case
        super().__init__()

        # Get config from files
        self.update(self._get_config_from_files())

        # Add override parameters
        if len(override) > 0:
            self.update(override)

        # Version type cleanup
        if not isinstance(self['pythonversion'], PythonVersion):
            self['pythonversion'] = PythonVersion.from_config(
                arch = self['arch'],
                version = self['pythonversion'],
            )

    def __getitem__(self, key: str) -> Any:
        """
        Returns values from the following sources in the following order:

        - Environment variables
        - Internal storage, i.e. changed in the dictionary or read from configuration files.
        - Default values.

        Args:
            key : Name of configuration value.
        Returns:
            Arbitrary configuration value.
        """

        env_var = "WENV_{NAME:s}".format(NAME=key.upper())
        if env_var in os.environ.keys():
            value = os.environ[env_var]
            if len(value) > 0:
                if key == "pythonversion":
                    return PythonVersion.from_config(self['arch'], value)
                if value.isnumeric():
                    return int(value)
                if value.strip().lower() in ("true", "false"):
                    return {"true": True, "false": False}[value.strip().lower()]
                return value

        if key in self.keys():
            return super().__getitem__(key)

        if key == "arch":
            return "win32"  # Define Wine & Wine-Python architecture
        if key == "pythonversion":
            return PythonVersion(self['arch'], 3, 7, 4, 'stable')  # Define Wine-Python version
        if key == "wine_bin_win32":
            return "wine"
        if key == "wine_bin_win64":
            return "wine64"
        if key == "wine_bin_arm64":
            return "wine"
        if key == "winedebug":
            return "-all"  # Wine debug output off
        if key == "wineinstallprefix":
            return None  # no custom Wine installation outside of PATH
        if key == "prefix":
            install_location = os.path.abspath(__file__)
            if install_location.startswith(
                site.USER_BASE
            ):  # Hacky way of looking for a user installation
                return site.USER_BASE
            return sys.prefix
        if key == "wineprefix":
            return os.path.join(self["prefix"], "share", "wenv", self["arch"])
        if key == "pythonprefix":
            return os.path.join(
                self["wineprefix"], "drive_c", "python-%s" % self["pythonversion"]
            )
        if key == "offline":
            return False
        if key == "cache":
            return os.path.join(self["prefix"], "share", "wenv", "cache")
        if key == "packages":
            return os.path.join(self["cache"], "packages")
        if key == "_issues_50_workaround":
            return False  # Workaround for zugbruecke issue #50 (symlinks ...)

        raise KeyError("not a valid configuration key", key)

    def export_dict(self) -> Dict[str, Any]:
        """
        Exports a dictionary.
        """

        return {field: self[field] for field in self._KEYS}

    def export_envvar_dict(self) -> Dict[str, str]:
        """
        Exports a dictionary which can passed to the OS as a set of environment variables for ``wenv`` itself.
        """

        return {
            "WENV_" + field.upper(): "" if field is None else str(self[field])
            for field in self._KEYS
        }

    def _get_config_from_files(self) -> Dict:

        base = {}

        # Look for config in the usual spots
        for fn in [
            "/etc/wenv",
            os.path.join("/etc", CONFIG_FN), # TODO deprecated
            os.path.join("/etc", CONFIG_FN[1:]),
            os.path.join(os.path.expanduser("~"), CONFIG_FN),
            os.environ.get("WENV"),
            os.path.join(os.environ.get("WENV"), CONFIG_FN)
            if os.environ.get("WENV") is not None
            else None,
            os.path.join(os.getcwd(), CONFIG_FN),
        ]:

            cnt = self._load_config_from_file(fn)

            if cnt is not None:
                base.update(cnt)

        return base

    def _load_config_from_file(self, try_path: Optional[str] = None) -> Optional[Dict[str, Any]]:

        # If there is a path ...
        if try_path is None:
            return

        # Is this a file?
        if not os.path.isfile(try_path):
            return

        # Read file
        try:
            with open(try_path, "r", encoding="utf-8") as f:
                cnt = f.read()
        except Exception as e:
            raise EnvConfigParserError(
                'Config file could not be read: "%s"' % try_path
            ) from e

        # Try to parse it
        try:
            cnt = json.loads(cnt)
        except Exception as e:
            raise EnvConfigParserError(
                'Config file could not be parsed: "%s"' % try_path
            ) from e

        # Ensure that config has the right format
        if not isinstance(cnt, dict):
            raise EnvConfigParserError('Config file is malformed: "%s"' % try_path)

        return cnt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    docs/source/conf.py: Configures the documentation build process

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

# sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))

from docs.source.version import get_version

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'sphinx_rtd_theme',
    'myst_parser',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "wenv"
author = "Sebastian M. Ernst"
copyright = f"2017-2022 {author:s}"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = get_version()
# The full version, including alpha/beta/rc tags.
release = get_version()

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Values to pass into the template engine's context for all pages.
html_context = {
    'sidebar_external_links_caption': 'Links',
    'sidebar_external_links': [
        # ('<i class="fa fa-rss fa-fw"></i> Blog', 'https://www.000'),
        ('<i class="fa fa-github fa-fw"></i> Source Code', 'https://github.com/pleiszenburg/wenv'),
        ('<i class="fa fa-bug fa-fw"></i> Issue Tracker', 'https://github.com/pleiszenburg/wenv/issues'),
        ('<i class="fa fa-envelope fa-fw"></i> Mailing List', 'https://groups.io/g/zugbruecke-dev'),
        ('<i class="fa fa-comments fa-fw"></i> Chat', 'https://matrix.to/#/#zugbruecke:matrix.org'),
        # ('<i class="fa fa-file-text fa-fw"></i> Citation', 'https://doi.org/000'),
        ('<i class="fa fa-info-circle fa-fw"></i> pleiszenburg.de', 'https://www.pleiszenburg.de/'),
    ],
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "relations.html",  # needs 'show_related': True theme option to display
        "searchbox.html",
        "donate.html",
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = f"{project:s}doc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, f"{project:s}.tex", f"{project:s} Documentation", author, "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, project, f"{project:s} Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        project,
        f"{project:s} Documentation",
        author,
        project,
        "One line description of project.",
        "Miscellaneous",
    ),
]


always_document_param_types = True # sphinx_autodoc_typehints

napoleon_include_special_with_doc = True # napoleon
# napoleon_use_param = True
# napoleon_type_aliases = True



# -*- coding: utf-8 -*-

"""
WENV
Running Python on Wine
https://github.com/pleiszenburg/wenv

    tests/test_util.py: Testing pip

    Copyright (C) 2017-2022 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/wenv/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from .lib import get_context, run_process, no_errors_in

from wenv import Env

import pytest

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# TEST(s)
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@pytest.mark.parametrize("arch,build", get_context())
def test_pip(arch, build):

    out, err, code = run_process(["wenv", "pip", "list"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})
    assert code == 0
    assert no_errors_in(err)
    assert "pip" in out
    assert "setuptools" in out
    assert "pytest" not in out

    out, err, code = run_process(
        ["wenv", "pip", "install", "pytest"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)}
    )
    assert code == 0
    assert no_errors_in(err)

    out, err, code = run_process(["wenv", "pip", "list"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})
    assert code == 0
    assert "pip" in out
    assert "setuptools" in out
    assert "pytest" in out
    assert no_errors_in(err)


@pytest.mark.parametrize("arch,build", get_context())
def test_pip_api(arch, build):

    env = Env(arch=arch, pythonversion=build)

    out, err, code = run_process(["wenv", "pip", "list"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})
    assert code == 0
    assert no_errors_in(err)
    assert "pip" in out
    assert "setuptools" in out
    assert "requests" not in out

    packages = [package["name"] for package in env.list_packages()]
    assert "pip" in packages
    assert "setuptools" in packages
    assert "requests" not in packages

    env.install_package("requests")

    out, err, code = run_process(["wenv", "pip", "list"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})
    assert code == 0
    assert no_errors_in(err)
    assert "pip" in out
    assert "setuptools" in out
    assert "requests" in out

    packages = [package["name"] for package in env.list_packages()]
    assert "pip" in packages
    assert "setuptools" in packages
    assert "requests" in packages

    env.uninstall_package("requests")

    out, err, code = run_process(["wenv", "pip", "list"], env={"WENV_ARCH": arch, "WENV_PYTHONVERSION": str(build)})
    assert code == 0
    assert no_errors_in(err)
    assert "pip" in out
    assert "setuptools" in out
    assert "requests" not in out

    packages = [package["name"] for package in env.list_packages()]
    assert "pip" in packages
    assert "setuptools" in packages
    assert "requests" not in packages
